from os.path import isfile
from datetime import datetime
from collections import defaultdict
from openpyxl import load_workbook
from openpyxl.worksheet.cell_range import CellRange
from openpyxl.utils.cell import coordinate_to_tuple
from openpyxl.cell import Cell
from . import log, version
from .utils import is_vertical_range
from .parser import Parser

import re
import yaml
import dictdiffer

ANON_PRFX = 'anon_'

def useless_range(rng):
    """
    Check if rng is valid and useful range.
    Empty range is not significative.

    :param rng:
    :return: True | False
    """
    if not rng:
        return True


class Collector:
    """
    Main Collector class
    """
    label_names = [re.compile('^varname', re.IGNORECASE),
                   re.compile('^label', re.IGNORECASE)]
    model_names = [re.compile('^model', re.IGNORECASE)]
    data_names = [re.compile('^data', re.IGNORECASE)]

    def __init__(self, url:str, only_data:bool=False, relative:bool=False, add_fingerprint:bool=False,
                 parsed:bool=False):
        """
        Data injestion, we need 2 instances of the sheet:
          - wb_data: with static data
          - wb: with the formulas

        :param url: url to an input excel file
        :param only_data: read only static values
        :param parsed: use parsed formaulas instead of excel version
        :param add_fingerprint: add some punctual information in the internal representation
        :param relative: all area are treated as if they starts from Row1 Col1
        """
        if not isfile(url):
            raise FileNotFoundError("File {} does not exists".format(url))

        self.sheets = {}
        self.pseudo = {}
        self.url = url
        self.anon_models = {}
        self.parsed = parsed

        self.labels_as_data = True
        self.models_as_data = only_data
        log.debug("Labels read as {}".format( 'data' if self.labels_as_data else 'formula'))
        log.debug("Models read as {}".format('data' if self.models_as_data else 'formula'))
        if self.parsed:
            log.info("Parser enabled")

        # workbook injestion, we need 2 instances: data and formulas
        # if only_data is true each point to the same instance.

        self.wb_data = load_workbook(filename=self.url, data_only=True)
        if only_data:
            self.wb = self.wb_data
        else:
            self.wb = load_workbook(filename=self.url)
        self.relative = relative
        self.add_fingerprint = add_fingerprint

        self.sheetnames = self.wb_data.sheetnames
        self.named_ranges = self.wb_data.defined_names.definedName
        self.set_ranges()

        self.labels = self.handle_range(
            self.label_names,
            self.labels_as_data)

        # Here we have collected only labels, so if parsed is True
        # we need a data structure for bind position of formula to his label
        self.sheet_is_vertical = {}
        if self.parsed:
            # models pre-scan, this solves the anonymous models problems.
            # we don't know if they are vertical or horizontal
            self.models = self.handle_range(
                self.model_names,
                True)

            self.find_anonymous_models()

            self.pos_to_label = defaultdict(lambda: {})
            for sheet,rng in self.labels.items():
                if sheet not in self.sheet_is_vertical:
                    is_vert = is_vertical_range(rng.keys())
                    self.sheet_is_vertical[sheet] = is_vert
                else:
                    is_vert = self.sheet_is_vertical[sheet]
                for coord, label in rng.items():
                    row, col = coordinate_to_tuple(coord)
                    self.pos_to_label[sheet][col if is_vert else row] = label

            self.parser = Parser(collector=self)

            self.models = self.handle_range(
                self.model_names,
                self.models_as_data)
        else:
            self.models = self.handle_range(
                            self.model_names,
                            self.models_as_data)

            self.find_anonymous_models()

        self.data = self.handle_range(
                      self.data_names,
                      use_data=True)

        self.set_pseudo_excel()



    def find_anonymous_models(self):
        """
        Models can be defined without labels, so in this cases we must do
        more to:
          - Assign default labels
          - Understand if model is vertical or horizontal

        :return:
        """
        # Check for anonymous models
        for k,v in self.models.items():
            if k not in self.labels:
                self.sheet_is_vertical[k] = is_vertical_range(v)
                anon_labels = {f'{x}': f'{ANON_PRFX}{n+1}' for n, x in enumerate(v)}
                self.anon_models[k] = '{} anon labels assigned'.format(len(anon_labels))
                self.labels[k] = anon_labels

        for k,v in self.anon_models.items():
            log.warning("Found anonymous model {} : {}".format(k,v))


    def set_ranges(self):
        """
        Iter over named ranges and collect ALL ranges into:
            - self.ranges
            - self.params
            - self.text_ranges

        here we dont know if ranges are used, but we cannot distinguish cell_alias, and other stuff.

        :return:
            None
        """
        self.ranges = {}
        self.text_ranges = {}
        self.params = {}

        for x in self.named_ranges:
            for sheet, rng in x.destinations:
                if x.type == 'TEXT':
                    self.text_ranges[x.name] = x.value
                elif x.type == 'NUMBER':
                    self.params[x.name] = x.value
                else:
                    if useless_range(rng):
                        log.debug("{} range [{}]in sheet {} discarded ".format(x.name,rng,sheet))
                        continue
                    try:
                        cells = self.wb_data[sheet][rng]
                    except KeyError as err:
                        log.error(err)
                        continue

                    if isinstance(cells,Cell):
                        # Single named cell
                        self.ranges[x.name] = (cells,)
                    else:
                        # Range named
                        self.ranges[x.name] = sum(cells, ())

        log.debug("{} named ranges collected".format(len(self.ranges)))
        log.debug("{} parameters collected".format(len(self.params)))
        log.debug("{} text range collected".format(len(self.text_ranges)))

    def to_relative(self, rng_name:str, cell:Cell) -> str:
        """
        given a range name and a cell inside it, it return che cell position
        as R<row>C<col> as relative to the cell 0,0

        :param rng_name: collection of cells
        :param cell: target cell
        :return: relative position of the cell.
        """
        ref_cell = self.ranges[rng_name][0]
        min_row, min_col = ref_cell.row, ref_cell.column
        return "R{0}C{0}".format(cell.row - min_row + 1, cell.column - min_col + 1)


    def handle_range(self, labels: list, use_data:bool):
        """
        iter over collected ranges and restructure all data in a confortable nested dict.
        coll[<sheet_name>] = {..}

        :param labels: list of RE to use to undestand type of range
        :param use_data: Use data or formulas.
        :return: collection of all collected cells
        """
        coll = {}
        for lbl in self.ranges:
            for check_valid in labels:
                if check_valid.match(lbl):
                    cells = self.ranges[lbl]
                    sheet_names = set([x.parent.title for x in cells])
                    if len(sheet_names) > 1:
                        raise NotImplementedError("Range defined on multiple sheets is not handled")
                    sheet_name = sheet_names.pop()
                    if use_data:
                        coll[sheet_name] = {self.to_relative(lbl,x)
                                            if self.relative
                                            else x.coordinate:x.value
                                            for x in cells}
                    else:
                        coll[sheet_name] = {self.to_relative(lbl,x)
                                            if self.relative
                                            else x.coordinate:self.load_formula(sheet_name,  x.coordinate,  self.wb[sheet_name][x.coordinate].value)
                                            for x in cells}
                    break
        return coll


    def load_formula(self, sheet:str, position:str, s:str) -> str:
        """
        In the case we need to handle a formula we need an optional parameter
        to activate the parser. *self.parsed*
        If parsed is True so we do a further pass for rewrite in pythonic way.

        :param sheet:  sheet containing cell
        :param coordinate: cell coordinate
        :param s: input formula a simple string
        :return: s | transliterated version
        """
        if (not self.parsed) or (not isinstance(s,str)):
            return s

        log.debug('Parsing {} {} {}'.format(sheet, position, s))
        # In case of anonymous sheet we must force
        self.parser.set_current(sheet, coordinate_to_tuple(position))
        ret = self.parser.transform(s)
        log.debug('Parsed {}'.format(ret))
        return ret



    def set_pseudo_excel(self):
        if not self.pseudo:
            # Remember dict are ordered!
            if self.add_fingerprint:
                self.pseudo['xltoy.version'] = version
                self.pseudo['xltoy.filename'] = self.url
                self.pseudo['xltoy.datetime'] = datetime.now().isoformat()

            for sheet, labels in self.labels.items():
                self.pseudo[sheet] = dict(zip(labels.values(), self.models[sheet].values()))
                if sheet in self.data:
                    self.pseudo[sheet]['data'] = self.data[sheet]

            # We must handle data ranges not touched before, here
            # we have only data range not in sheet with labels
            for sheet in set(self.data) - set(self.labels):
                self.pseudo[sheet] = self.data[sheet]

    def to_yaml(self):
        self.set_pseudo_excel()
        return yaml.dump(self.pseudo)


class YamlCollector(Collector):
    """
    Collector for yaml format, it is used only to set
    self.pseudo attribute.
    """
    def __init__(self, url):
        if not isfile(url):
            raise FileNotFoundError("File {} does not exists".format(url))

        with open(url) as fin:
            self.pseudo = yaml.load(fin, Loader=yaml.FullLoader)



class DiffCollector:
    def __init__(self, url1, url2, only_data:bool=False, relative:bool=False, add_fingerprint:bool=False,
                 parsed:bool=False):
        """
        workbook differ, given 2 files (excel, yaml or json) it can do intelligent comparison

        :param url1:
        :param url2:
        :param only_data: ignore formulas and compare only values
        :param relative: all area are treated as if they starts from Row1 Col1
        :param parsed: use parsed formaulas instead of excel version
        :param add_fingerprint:
        """
        c1,c2 = [YamlCollector(url)
                 if url.lower().endswith('yaml')
                 else Collector(url,only_data=only_data, relative=relative,
                                    add_fingerprint=add_fingerprint, parsed=parsed)
                 for url in [url1,url2]]

        self.iter_differs = dictdiffer.diff(c1.pseudo,c2.pseudo)
        self.do_diff()

    def do_diff(self):
        self.diff = {}
        for kind, mid, sh_cells in self.iter_differs:
            if kind not in self.diff:
                self.diff[kind] = {}
            if isinstance(mid, list):
                if len(mid) == 1:
                    mid = mid[0]
                else:
                    raise RuntimeError("{} not understood".format(mid))
            if kind == 'change':
                sheet, label = mid.split('.')
                if sheet not in self.diff[kind]:
                    self.diff[kind][sheet] = {}
                self.diff[kind][sheet][label] = ' -> '.join(['{}'.format(x) for x in sh_cells])
            else:
                for sheet, cells in sh_cells:
                    if sheet not in self.diff[kind]:
                        self.diff[kind][sheet] = {}
                    self.diff[kind][sheet] = cells

    def to_yaml(self):
        if self.diff:
            print(yaml.dump(self.diff))




# -*- coding: utf-8
"""
xltoy main cli (Command Line Interface)
"""
import click
from xltoy.collector import Collector, DiffCollector
from xltoy.utils import timeit
from xltoy import *

def set_verb(v=0):
    """

    :param v:  verbosity 1:WARNING,
                         2:INFO,
                         3+:DEBUG
    :return: Nothing,but logging verbosisty was set
    """
    if v>2:
        log.setLevel(DEBUG)
    elif v>1:
        log.setLevel(INFO)
    else:
        log.setLevel(WARNING)

@click.group()
def cli():
    pass

@click.command()
@click.option('--timeit', is_flag=True, help='Print out how many times it takes for the task')
@click.option('--yaml', is_flag=True, help='Print out the yaml hiearchical view')
@click.option('--data', is_flag=True, help='Collect only data, it will ignore formulas')
@click.option('-v', '--verbose', count=True, help="verbose output (repeat for increased verbosity)")
@click.option('--add_fingerprint', is_flag=True, help='Add metadata under section xltoy')
@click.option('--parsed', is_flag=True, help='Parse formulas and use this version instead of excel syntax')
@click.argument('filename')
def collect(filename, **kwargs):
    set_verb(kwargs.get('verbose'))
    with timeit("{} collect".format(filename), kwargs.get('timeit')):
        c = Collector(filename,
                      only_data=kwargs.get('data'),
                      parsed=kwargs.get('parsed'),
                      add_fingerprint=kwargs.get('add_fingerprint'),
                      )
        if kwargs.get('yaml'):
            print(c.to_yaml())


@click.command()
@click.option('--timeit', is_flag=True, help='Print out how many times it takes for the task')
@click.option('--data', is_flag=True, help='Collect only data, it will ignore formulas')
@click.option('--relative', is_flag=True, help='Areas are handled as relative, each starts from row1,col1')
@click.option('-v', '--verbose', count=True, help="verbose output (repeat for increased verbosity)")
@click.option('--add_fingerprint', is_flag=True, help='Add metadata under section xltoy')
@click.option('--parsed', is_flag=True, help='Parse formulas and use this version instead of excel syntax')
@click.argument('filename1')
@click.argument('filename2')
def diff(filename1, filename2, **kwargs):
    set_verb(kwargs.get('verbose'))
    with timeit("collect 2 files", kwargs.get('timeit')):
        d = DiffCollector(filename1,filename2,
                          only_data=kwargs.get('data'),
                          relative=kwargs.get('relative'),
                          parsed=kwargs.get('parsed'),
                          add_fingerprint=kwargs.get('add_fingerprint'))
        d.to_yaml()

cli.add_command(collect)
cli.add_command(diff)

if __name__ == '__main__':
    cli()



import xltoy
from xltoy.collector import Collector, DiffCollector
from xltoy.utils import is_vertical_range
import os
import pytest

base_data_url = os.path.join(os.path.dirname(os.path.dirname(xltoy.__file__)),'data')



def coll(url):
    url=os.path.join(base_data_url,url)
    return Collector(url)


@pytest.mark.collector
@pytest.mark.parametrize("f", [src for src in list(os.walk(base_data_url))[0][2] if src.endswith('.xlsx')])
def test_injestion(f):
    coll(f)


@pytest.mark.collector
@pytest.mark.parametrize("rng", [['A1','A2','A3','$A$4','$A5','A$12','$A$13']])
def test_is_vertical(rng):
    assert is_vertical_range(rng) is False


@pytest.mark.collector
@pytest.mark.parametrize("rng", [['AA1','AA2','AA3','$AA5','$AA$6']])
def test_is_horiz(rng):
    assert is_vertical_range(rng) is False


@pytest.mark.collector
@pytest.mark.parametrize("rng",[['A1','B2','C3']])
def test_not_valid_range(rng):
    with pytest.raises(ValueError):
        is_vertical_range(rng) is True

@pytest.mark.collector
@pytest.mark.parametrize("rng",[])
def test_empty_valid_range(rng):
    with pytest.raises(ValueError):
        is_vertical_range(rng)

@pytest.mark.collector
@pytest.mark.parametrize("f", [src for src in list(os.walk(base_data_url))[0][2] if src.endswith('.xlsx')])
def test_diff(f):
    f = os.path.join(base_data_url, f)
    stone_f = os.path.join(
                  base_data_url,
                  os.path.splitext(f)[0]+'.yaml')
    if os.path.exists(stone_f):
        assert not DiffCollector(f, stone_f).diff

    stone_f = os.path.join(
                  base_data_url,
                  os.path.splitext(f)[0]+'.parsed.yaml')
    if os.path.exists(stone_f):
        assert not DiffCollector(f, stone_f, parsed=True).diff



import pytest
from xltoy.parser import Parser
from xltoy.utils import split_sheet_coordinates

p = Parser(collector=None)

test_set = """
=3*A7+5
=+3
=3*Sheet1!$A$7+5
=3*'Sheet 1'!$A$7+5
=3*'O''Reilly''s sheet'!$A$7+5
=if(Sum(A1:A25)>42,Min(B1:B25),if(Sum(C1:C25)>3.14, (Min(C1:C25)+3)*18,Max(B1:B25)))
=sum(a1:a25,10,min(b1,c2,d3))
=if("T"&a2="TTime", "Ready", "Not ready")
=E3
=+8
=E4+Y6+AA3
=(E3+D3)/2
=+LOG(F11)
=(E4+D4)/2
=IF(D7,D7+E7,D7-E7)
=1/(1+EXP(LN(1/INPUT_M!$G$62-1)-E$543))
=+IF(A4+B4, 1/(1+EXP(LN(1/INPUT_M!$G$62-1)-E$543)), IF(A4+B4, 1, PRIVFUN(E4)))
=+IF(PARAM!$B$135=0, 1/(1+EXP(LN(1/INPUT_M!$G$62-1)-E$543)), IF(PARAM!$B$135=1,PRIVFUN(FUNCT(INPUT_M!$G$62)+E$543),PRIVFUN((FUNCT(INPUT_M!$G$62)+SQRT(INPUT_M!$G$45)*E$543)/SQRT(1-INPUT_MATRIX!$G$45))))
=K13-MIN(L$10,0)+IF(SHEE1!M20=1,0,MIN(K10,0))+L15+L16
=+INTERBANK!FS221
""".splitlines()

@pytest.mark.parser
@pytest.mark.parametrize("s", filter(None,test_set))
def test_parser(s):
    assert p.parse(s)


@pytest.mark.parser
def test_split_sheet_coordinates():
    assert split_sheet_coordinates('E4')          == (None, 'E4')
    assert split_sheet_coordinates('MYSH!E4')     == ('MYSH','E4')
    assert split_sheet_coordinates("'M Y SH'!E4") == ('M Y SH', 'E4')




import logging
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL, Formatter, getLogger, StreamHandler
__version__ = '0.1.6'
version = __version__


default_log_name = 'xltoy'
default_log_level = ERROR
default_log_format = Formatter('%(levelname).3s %(message)s')
log = getLogger(default_log_name)
log.setLevel(default_log_level)
str_log = logging.StreamHandler()
str_log.setFormatter(default_log_format)
log.addHandler(str_log)





import setuptools
from xltoy import version

with open("README.md", "rt") as fh:
    long_description = fh.read()

with open("requirements.txt",'rt') as fn:
    reqs = fn.read().split()

setuptools.setup(
    name="xltoy",
    version=version,
    author="Glauco Uri",
    author_email="glauco@uriland.it",
    description="The ultimate toolkit for Microsoft Excel modelers and mod-operations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/glaucouri/xltoy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
    ],
    python_requires='>=3.6',
    install_requires = reqs,
    entry_points={
            'console_scripts': [
                'xltoy = scripts.cmd:cli'
          ]}
)


from contextlib import contextmanager
from openpyxl.cell.cell import Cell
from openpyxl.utils.cell import coordinate_to_tuple
from openpyxl.workbook.defined_name import SHEETRANGE_RE
from time import time


@contextmanager
def timeit(description: str, enabled:bool=False) -> None:
    """
    Generic Context manager  simple Timeit

    :param description:
    :param enabled:
    :return:
    """
    t0 = time()
    yield
    ellapsed_time = time() - t0

    if enabled:
        print(f"{description} done in {ellapsed_time:.3f} s.")


def de_dollar(s:str)-> str:
    """
    Remove useless dollars

    :param s: string representing a coordinate
    :return: same cell cleanup

    de_dollar('$AA$4') -> 'AA4

    """
    return s.replace('$','')

def is_vertical_range(rng:list)->bool:
    """

    :param rng:
    :return:
    """
    rows_cols = [coordinate_to_tuple(de_dollar(x)) for x in rng]
    rows = set((x[0] for x in rows_cols))
    cols = set((x[1] for x in rows_cols))
    if len(rows)==1:
        if len(cols)>1:
            return True
        else:
            raise ValueError("Cannot recognize if {} is vertical model range")
    else:
        # We assume it was >1
        if len(cols) == 1:
            return False
        else:
            raise ValueError("Cannot recognize if {} is horizontal model range")


def split_sheet_coordinates(s):
    """

    :param s:
    :return:
    """
    match = SHEETRANGE_RE.match(s)
    if match is None:
        return None, s
    else:
        match = match.groupdict()
        return match['quoted'] or match['notquoted'] , match['cells']



# inspired to: excelExpr by Paul McGuire
from openpyxl.utils.cell import coordinate_to_tuple
from . import log
from .utils import de_dollar
from pyparsing import (
    CaselessKeyword, Word, alphas, alphanums, nums, Optional, Group, oneOf, Forward,
    infixNotation, opAssoc, dblQuotedString, delimitedList, Combine, Literal, QuotedString, ParserElement,
    LineEnd, pyparsing_common as ppc)
ParserElement.enablePackrat()





class Parser:
    def __init__(self, collector=None):
        self.current_sheet = None
        self.collector = collector
        EQ, LPAR, RPAR, COLON, COMMA, EXCL, DOLLAR = map(Literal, "=():,!$")
        multOp = oneOf("* /")
        addOp = oneOf("+ -")
        words = Word(alphas, alphanums + '_')
        sheetRef = words | QuotedString("'", escQuote="''")
        colRef = Optional(DOLLAR) + Word(alphas, max=2)
        rowRef = Optional(DOLLAR) + Word(nums)

        cellRef = Combine(
            Group(Optional(sheetRef("sheet") + EXCL) + colRef("col") + rowRef("row"))
        ).setParseAction(self.cell_action)

        cellRange = (
                Group(cellRef("start") + COLON + cellRef("end"))("range")
                | cellRef
                | Word(alphas, alphanums)
        )('cells')

        expr = Forward()

        COMPARISON_OP = oneOf("< = > >= <= != <>")
        condExpr = expr + Optional(COMPARISON_OP + expr)

        ifFunc = (
                CaselessKeyword("if")
                + LPAR
                + Group(condExpr)("condition")
                + COMMA
                + Group(expr)("if_true")
                + COMMA
                + Group(expr)("if_false")
                + RPAR
        )

        def stat_function(name):
            return Group(CaselessKeyword(name) + Group(LPAR + delimitedList(expr, combine=True) + RPAR))

        sumFunc = stat_function("sum")
        minFunc = stat_function("min")
        maxFunc = stat_function("max")
        aveFunc = stat_function("ave")
        sqrFunc = stat_function("sqrt")
        unknowFunc = words + Group(LPAR + expr + RPAR)

        funcCall = ifFunc | sumFunc | minFunc | maxFunc | aveFunc | sqrFunc | unknowFunc

        numericLiteral = ppc.number
        operand = numericLiteral | funcCall | cellRange | cellRef
        arithExpr = infixNotation(
            operand, [(multOp, 2, opAssoc.LEFT),
                      (addOp, 2, opAssoc.LEFT)],
            lpar=LPAR, rpar=RPAR
        )

        textOperand = dblQuotedString | cellRef
        textExpr = infixNotation(textOperand, [("&", 2, opAssoc.LEFT), ])

        atom = (arithExpr | textExpr)
        expr << Optional(addOp) + atom
        bnf = Optional(EQ|addOp) + expr + LineEnd()
        self.bnf = bnf

    def transform(self, *args, **kwargs):
        """
        a simple hook to transformString function from pyparsing

        :param args: delegated to child function
        :param kwargs: delegated to child function
        :return: delegated to child function
        """
        return self.bnf.transformString(*args, **kwargs)

    def parse(self, *args, **kwargs):
        return self.bnf.parseString(*args, **kwargs)

    def set_current(self, sheet, coordinate):
        self.current_sheet = sheet
        self.current_row, self.current_col = coordinate
        self.current_pos_to_label = self.collector.pos_to_label[sheet]
        self.current_sheet_is_vertical = self.collector.sheet_is_vertical[sheet]

    def cell_action(self, tok):
        """
        over all token occurrence we do transliteration action!

        tok is a cell reference, in these cases we can transliterate syntax in a
        python way.
        Depending on flag self.current_sheet_is_vertical we choose the right
        direction for the time stepper

        :param tok: exel token like E5, $AA$12
        :return: transliterated syntax
        """
        if self.collector is not None:
            if not self.current_sheet:
                raise ValueError("ggg")

            tok = tok[0]
            if '!' in tok:
                # maybe from other sheet!
                sheet, f_tok = tok.split('!')
                if sheet != self.current_sheet:
                    log.warning(f"In sheet {self.current_sheet} cell {tok} found input from external sheet. Treated as exogenous value")
                    val = self.collector.wb_data[sheet][f_tok].value
                    return val

            tok = de_dollar(tok)

            row, col = coordinate_to_tuple(tok)
            pos = col if self.current_sheet_is_vertical else row
            if pos in self.current_pos_to_label:
                label = self.current_pos_to_label[pos]
            else:
                val = self.collector.wb_data[self.current_sheet][tok].value
                log.info("Found params sheet {} cell {}:{}".format(self.current_sheet,tok,val))
                return val

            delta_time = row-self.current_row if self.current_sheet_is_vertical else col-self.current_col
            if delta_time:
                return("{}[T{}]".format(label,delta_time))
            else:
                return("{}[T]".format(label))
